@@ -1,11 +1,14 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE LambdaCase #-}
 -- |
--- Stability: experimental
+-- Stability: unstable
 --
--- This module contains formatters that can be used with
--- `Test.Hspec.Core.Runner.hspecWith`.
-module Test.Hspec.Core.Formatters.V2 (
-
+-- This is an unstable API.  Use
+-- [Test.Hspec.Api.Formatters.V3](https://hackage.haskell.org/package/hspec-api/docs/Test-Hspec-Api-Formatters-V3.html)
+-- instead.
+module Test.Hspec.Core.Formatters.V2
+-- {-# WARNING "Use [Test.Hspec.Api.Formatters.V3](https://hackage.haskell.org/package/hspec-api/docs/Test-Hspec-Api-Formatters-V3.html) instead." #-}
+(
 -- * Formatters
   silent
 , checks
@@ -21,12 +24,20 @@ module Test.Hspec.Core.Formatters.V2 (
 -- Actions live in the `FormatM` monad.  It provides access to the runner state
 -- and primitives for appending to the generated report.
 , Formatter (..)
+, Path
+, Progress
+, Location(..)
 , Item(..)
 , Result(..)
 , FailureReason (..)
 , FormatM
 , formatterToFormat
 
+-- ** Accessing config values
+, getConfig
+, getConfigValue
+, FormatConfig(..)
+
 -- ** Accessing the runner state
 , getSuccessCount
 , getPendingCount
@@ -58,34 +69,37 @@ module Test.Hspec.Core.Formatters.V2 (
 , outputUnicode
 
 , useDiff
+, diffContext
+, externalDiffAction
 , prettyPrint
 , prettyPrintFunction
 , extraChunk
 , missingChunk
 
+-- ** expert mode
+, unlessExpert
+
 -- ** Helpers
 , formatLocation
-, formatException
+, Util.formatException
 
 #ifdef TEST
 , Chunk(..)
 , ColorChunk(..)
-, indentChunks
 #endif
 ) where
 
 import           Prelude ()
 import           Test.Hspec.Core.Compat hiding (First)
+import           System.IO (hFlush, stdout)
 
-import           Data.Char
-import           Data.Maybe
-import           Test.Hspec.Core.Util
+import           Test.Hspec.Core.Util hiding (formatException)
+import qualified Test.Hspec.Core.Util as Util
 import           Test.Hspec.Core.Clock
-import           Test.Hspec.Core.Example (Location(..))
+import           Test.Hspec.Core.Example (Location(..), Progress)
 import           Text.Printf
 import           Test.Hspec.Core.Formatters.Pretty.Unicode (ushow)
 import           Control.Monad.IO.Class
-import           Control.Exception
 
 -- We use an explicit import list for "Test.Hspec.Formatters.Monad", to make
 -- sure, that we only use the public API to implement formatters.
@@ -100,6 +114,10 @@ import Test.Hspec.Core.Formatters.Internal (
   , FormatM
   , formatterToFormat
 
+  , getConfig
+  , getConfigValue
+  , FormatConfig(..)
+
   , getSuccessCount
   , getPendingCount
   , getFailCount
@@ -126,29 +144,33 @@ import Test.Hspec.Core.Formatters.Internal (
   , outputUnicode
 
   , useDiff
+  , diffContext
+  , externalDiffAction
   , prettyPrint
   , prettyPrintFunction
   , extraChunk
   , missingChunk
+
+  , unlessExpert
   )
 
 import           Test.Hspec.Core.Formatters.Diff
 
 silent :: Formatter
 silent = Formatter {
-  formatterStarted      = return ()
-, formatterGroupStarted = \ _ -> return ()
-, formatterGroupDone    = \ _ -> return ()
-, formatterProgress     = \ _ _ -> return ()
-, formatterItemStarted  = \ _ -> return ()
-, formatterItemDone     = \ _ _ -> return ()
-, formatterDone         = return ()
+  formatterStarted      = pass
+, formatterGroupStarted = \ _ -> pass
+, formatterGroupDone    = \ _ -> pass
+, formatterProgress     = \ _ _ -> pass
+, formatterItemStarted  = \ _ -> pass
+, formatterItemDone     = \ _ _ -> pass
+, formatterDone         = pass
 }
 
 checks :: Formatter
 checks = specdoc {
   formatterProgress = \(nesting, requirement) p -> do
-    writeTransient $ indentationFor nesting ++ requirement ++ " [" ++ (formatProgress p) ++ "]"
+    writeTransient $ indentationFor nesting ++ requirement ++ " [" ++ formatProgress p ++ "]"
 
 , formatterItemStarted = \(nesting, requirement) -> do
     writeTransient $ indentationFor nesting ++ requirement ++ " [ ]"
@@ -161,10 +183,10 @@ checks = specdoc {
       Pending {} -> (withPendingColor, fallback "‐" "-")
       Failure {} -> (withFailColor,    fallback "✘" "x")
     case itemResult item of
-      Success {} -> return ()
-      Failure {} -> return ()
+      Success {} -> pass
+      Failure {} -> pass
       Pending _ reason -> withPendingColor $ do
-        writeLine $ indentationFor ("" : nesting) ++ "# PENDING: " ++ fromMaybe "No reason given" reason
+        indentBy (indentationFor ("" : nesting)) $ "# PENDING: " ++ fromMaybe "No reason given" reason
 } where
     indentationFor nesting = replicate (length nesting * 2) ' '
 
@@ -174,8 +196,7 @@ checks = specdoc {
       write $ indentationFor nesting ++ requirement ++ " ["
       withColor $ write symbol
       writeLine $ "]" ++ if shouldPrintTimes then times else ""
-      forM_ (lines info) $ \ s ->
-        writeLine $ indentationFor ("" : nesting) ++ s
+      indentBy (indentationFor ("" : nesting)) info
       where
         dt :: Int
         dt = toMilliseconds duration
@@ -209,7 +230,7 @@ specdoc = silent {
         writeResult nesting requirement duration info
       Pending _ reason -> withPendingColor $ do
         writeResult nesting requirement duration info
-        writeLine $ indentationFor ("" : nesting) ++ "# PENDING: " ++ fromMaybe "No reason given" reason
+        indentBy (indentationFor ("" : nesting)) $ "# PENDING: " ++ fromMaybe "No reason given" reason
       Failure {} -> withFailColor $ do
         n <- getFailCount
         writeResult nesting (requirement ++ " FAILED [" ++ show n ++ "]") duration info
@@ -221,8 +242,7 @@ specdoc = silent {
     writeResult nesting requirement (Seconds duration) info = do
       shouldPrintTimes <- printTimes
       writeLine $ indentationFor nesting ++ requirement ++ if shouldPrintTimes then times else ""
-      forM_ (lines info) $ \ s ->
-        writeLine $ indentationFor ("" : nesting) ++ s
+      indentBy (indentationFor ("" : nesting)) info
       where
         dt :: Int
         dt = floor (duration * 1000)
@@ -237,10 +257,12 @@ specdoc = silent {
 
 progress :: Formatter
 progress = failed_examples {
-  formatterItemDone = \ _ item -> case itemResult item of
-    Success{} -> withSuccessColor $ write "."
-    Pending{} -> withPendingColor $ write "."
-    Failure{} -> withFailColor $ write "F"
+  formatterItemDone = \ _ item -> do
+    case itemResult item of
+      Success{} -> withSuccessColor $ write "."
+      Pending{} -> withPendingColor $ write "."
+      Failure{} -> withFailColor $ write "F"
+    liftIO $ hFlush stdout
 }
 
 failed_examples :: Formatter
@@ -273,8 +295,9 @@ defaultFailedFormatter = do
       write ("  " ++ show n ++ ") ")
       writeLine (formatRequirement path)
       case reason of
-        NoReason -> return ()
+        NoReason -> pass
         Reason err -> withFailColor $ indent err
+        ColorizedReason err -> indent err
         ExpectedButGot preface expected_ actual_ -> do
           pretty <- prettyPrintFunction
           let
@@ -288,16 +311,26 @@ defaultFailedFormatter = do
 
           let threshold = 2 :: Seconds
 
-          mchunks <- liftIO $ if b
-            then timeout threshold (evaluate $ diff expected actual)
-            else return Nothing
 
-          case mchunks of
-            Just chunks -> do
-              writeDiff chunks extraChunk missingChunk
+          mExternalDiff <- externalDiffAction
+
+          case mExternalDiff of
+            Just externalDiff -> do
+              liftIO $ externalDiff expected actual
+
             Nothing -> do
-              writeDiff [First expected, Second actual] write write
+              context <- diffContext
+              mchunks <- liftIO $ if b
+                then timeout threshold (evaluate $ lineDiff context expected actual)
+                else return Nothing
+
+              case mchunks of
+                Just chunks -> do
+                  writeDiff chunks extraChunk missingChunk
+                Nothing -> do
+                  writeDiff [LinesFirst (splitLines expected), LinesSecond (splitLines actual)] write write
           where
+            writeDiff :: [LineDiff] -> (String -> FormatM ()) -> (String -> FormatM ()) -> FormatM ()
             writeDiff chunks extra missing = do
               writeChunks "expected: " (expectedChunks chunks) extra
               writeChunks " but got: " (actualChunks chunks) missing
@@ -305,72 +338,76 @@ defaultFailedFormatter = do
             writeChunks :: String -> [Chunk] -> (String -> FormatM ()) -> FormatM ()
             writeChunks pre chunks colorize = do
               withFailColor $ write (indentation ++ pre)
-              forM_ (indentChunks indentation_ chunks) $ \ chunk -> case chunk of
-                PlainChunk a -> write a
-                ColorChunk a -> colorize a
-              writeLine ""
+              go pass chunks
               where
+                indentation_ :: [Char]
                 indentation_ = indentation ++ replicate (length pre) ' '
 
+                go :: FormatM () -> [Chunk] -> FormatM ()
+                go indent_ = \ case
+                  [] -> pass
+                  c : cs -> do
+                    indent_
+                    case c of
+                      Original a -> write a
+                      Modified a -> colorize a
+                      Info text -> withInfoColor $ write text
+                      ModifiedChunks xs -> for_ xs $ \ case
+                        PlainChunk a -> write a
+                        ColorChunk a -> colorize a
+                    write "\n"
+                    go (write indentation_) cs
+
         Error info e -> do
           mapM_ indent info
+          formatException <- getConfigValue formatConfigFormatException
           withFailColor . indent $ "uncaught exception: " ++ formatException e
 
-      writeLine ""
 
-      let path_ = (if unicode then ushow else show) (joinPath path)
-      writeLine ("  To rerun use: --match " ++ path_)
+      unlessExpert $ do
+        let path_ = (if unicode then ushow else show) (joinPath path)
+        writeLine ""
+        seed <- usedSeed
+        writeLine ("  To rerun use: --match " ++ path_ <> " --seed " <> show seed)
       where
         indentation = "       "
-        indent message = do
-          forM_ (lines message) $ \line -> do
-            writeLine (indentation ++ line)
+        indent = indentBy indentation
 
-data Chunk = Original String | Modified String
+indentBy :: String -> String -> FormatM ()
+indentBy indentation message = do
+  forM_ (lines message) $ \ line -> do
+    writeLine (indentation ++ line)
+
+data Chunk = Original String | Modified String | Info String | ModifiedChunks [ColorChunk]
   deriving (Eq, Show)
 
-expectedChunks :: [Diff] -> [Chunk]
-expectedChunks = mapMaybe $ \ chunk -> case chunk of
-  Both a -> Just $ Original a
-  First a -> Just $ Modified a
-  Second _ -> Nothing
+expectedChunks :: [LineDiff] -> [Chunk]
+expectedChunks = concatMap $ \ case
+  LinesBoth a -> map Original a
+  LinesFirst a -> map Modified a
+  LinesSecond _ -> []
+  LinesOmitted n -> [Info $ formatOmittedLines n]
+  SingleLineDiff diffs -> return . ModifiedChunks . flip mapMaybe diffs $ \ case
+    First a -> Just $ ColorChunk a
+    Second _ -> Nothing
+    Both a -> Just $ PlainChunk a
 
-actualChunks :: [Diff] -> [Chunk]
-actualChunks = mapMaybe $ \ chunk -> case chunk of
-  Both a -> Just $ Original a
-  First _ -> Nothing
-  Second a -> Just $ Modified a
+actualChunks :: [LineDiff] -> [Chunk]
+actualChunks = concatMap $ \ case
+  LinesBoth a -> map Original a
+  LinesFirst _ -> []
+  LinesSecond a -> map Modified a
+  LinesOmitted n -> [Info $ formatOmittedLines n]
+  SingleLineDiff diffs -> return . ModifiedChunks . flip mapMaybe diffs $ \ case
+    First _ -> Nothing
+    Second a -> Just $ ColorChunk a
+    Both a -> Just $ PlainChunk a
 
 data ColorChunk = PlainChunk String | ColorChunk String
   deriving (Eq, Show)
 
-indentChunks :: String -> [Chunk] -> [ColorChunk]
-indentChunks indentation = concatMap $ \ chunk -> case chunk of
-  Original y -> [indentOriginal indentation y]
-  Modified y -> indentModified indentation y
-
-indentOriginal :: String -> String -> ColorChunk
-indentOriginal indentation = PlainChunk . go
-  where
-    go text = case break (== '\n') text of
-      (xs, _ : ys) -> xs ++ "\n" ++ indentation ++ go ys
-      (xs, "") -> xs
-
-indentModified :: String -> String -> [ColorChunk]
-indentModified indentation = go
-  where
-    go text = case text of
-      "\n" -> [PlainChunk "\n", ColorChunk indentation]
-      '\n' : ys@('\n' : _) -> PlainChunk "\n" : ColorChunk indentation : go ys
-      _ -> case break (== '\n') text of
-        (xs, _ : ys) -> segment xs ++ PlainChunk ('\n' : indentation) : go ys
-        (xs, "") -> segment xs
-
-    segment xs = case span isSpace $ reverse xs of
-      ("", "") -> []
-      ("", _) -> [ColorChunk xs]
-      (_, "") -> [ColorChunk xs]
-      (ys, zs) -> [ColorChunk (reverse zs), ColorChunk (reverse ys)]
+formatOmittedLines :: Int -> String
+formatOmittedLines n = "@@ " <> show n <> " lines omitted @@"
 
 defaultFooter :: FormatM ()
 defaultFooter = do
@@ -388,10 +425,12 @@ defaultFooter = do
          pluralize total   "example"
       ++ ", " ++ pluralize fails "failure"
       ++ if pending == 0 then "" else ", " ++ show pending ++ " pending"
-    c | fails /= 0   = withFailColor
+
+    color
+      | fails /= 0   = withFailColor
       | pending /= 0 = withPendingColor
       | otherwise    = withSuccessColor
-  c $ writeLine output
+  color $ writeLine output
 
 formatLocation :: Location -> String
 formatLocation (Location file line column) = file ++ ":" ++ show line ++ ":" ++ show column ++ ": "
